/* 
Calculations for HMM's = Hidden Markov Models

	bool MatrixInvert( Array2D< T > A ) - Matrix inversion by Gauss Jordan elimination 

	HMM :: CModelMath
			Provide a 'Q' model in the form of a TNT::Array2D then extract 
			various results as needed.  
		Array2D< T > m_aQ;		
		Array2D< T > m_aEigVec;	
		Array1D< T > m_aEigVal;	
		Array2D< T > m_aA;		
		Array1D< T > m_aPEq;	
		T m_dSampling;			
		CModelMath( iDim ) - Create the array 
		OutputDebug() - Write class contents to debug output window
		SetQElem( int i, int j, T dVal ) - Set 1 element of Q 
		SetDiagonal() - Set diagonal elements of Q to -1 * sum of row.	
		CalcEigen( ) - calculate eigenvalues and eigenvectors : m_aEigVec, m_aEigVal
		CalcA( T dSampling ) - Calculate A matrix 
		Consider : GetQ( ... GetEigVec( ...  GetEigVal( ...  GetA( ...  GetPEq( .. 

*/

#ifndef HMM_H
#define HMM_H

#include "JAMA_SVD.h"
#include "JAMA_eig.h"

using namespace TNT;

//===================================================================


//----------------------------------------------------------------------------
namespace HMM {
	const int MDL_DIAGONAL = 2;	// bit 1 = diagonal has been set
	const int MDL_EIGENVAL = 4;	// bit 2 = eigenvalues have been calculated 
	const int MDL_AMATRIX = 8;	// bit 3 = A Matrix and PEq have been calculated 
		
	template <class T>
	class CModelMath { 
		protected : 
			int m_iStatus;		// bit field, set to 0 when any inputs change.
			
		public : 
			/* 
			These items are public for easy access but should not be modified.
			If any are modified, the status variable needs to be set to 0.
			*/
			Array2D< T > m_aQ;			// Q matrix - rates 
			Array2D< T > m_aEigVec;		// eigenvectors
			Array1D< T > m_aEigVal;		// eigenvalues 
			Array2D< T > m_aA;			// exp(qt) = A matrix 
			Array1D< T > m_aPEq;		// Equilibrium probabilities
			T m_dSampling;				// Sampling used to calculate m_aA and m_aPEq
							
		public : 
			//-----					
			CModelMath( int iDim ) : m_aQ(iDim,iDim,0.0) { 
				m_iStatus=MDL_DIAGONAL;
				}

			CModelMath( Array2D<T> a ) : m_aQ(a) { 
				m_iStatus=MDL_DIAGONAL;
				}

			//-----
			void SetQElem( int i, int j, T dVal ) { 
				if( m_iStatus&MDL_DIAGONAL )	// if diagonal already set, update it
					m_aQ[i][i]+=m_aQ[i][j]-dVal;
				m_aQ[i][j]=dVal;				// change one Q Element 
				m_iStatus &= MDL_DIAGONAL;		// clear most of the flags 
				}

			//-----					
			void SetDiagonal() { 
				T dSum;
				for( int i=0; i<m_aQ.dim1(); ++i ) {
					dSum = 0.0;
					for( int j=0; j<m_aQ.dim2(); ++j )
						if( i!=j ) 
							dSum += m_aQ[i][j];
					m_aQ[i][i] = -1.0 * dSum;
					}
				m_iStatus |= MDL_DIAGONAL ;
				}
								
			//-----
			void CalcEigen( ) { 
				int i, j, iDim = m_aQ.dim1();
			
				//-- aQ = m_aQ * -1
				Array2D< T > aQ = m_aQ.copy();
				for( i=0; i<iDim; ++i ) 
					for( j=0; j<iDim; ++j ) 
						aQ[i][j] *= -1;

				//-- Calculate Eigenvalues
				JAMA::Eigenvalue<T> oQEigVal( aQ );

				//-- copy Eigenvalues 						
				m_aEigVal = Array1D< T > ( iDim );
				oQEigVal.getRealEigenvalues( m_aEigVal );
				
				//-- Sort for oQEigVal ascending - Initialize map of where sort ends up 
				TNT::Array1D< int > aMap(iDim);
				for( i=0; i<iDim; ++i) 
					aMap[i]=i;

				//-- sort eigvals and map 
				int iTmp, iState;
				T dTmp;
				for( iState=1; iState<iDim; ++iState ) { 
					dTmp = m_aEigVal[i=iState];
					iTmp = aMap[i];
					while( i>0 && m_aEigVal[i-1]>dTmp ) {
						aMap[i]=aMap[i-1];
						m_aEigVal[i]=m_aEigVal[i-1];
						--i;
						}
					if( i<iState ) {
						aMap[i]=iTmp;
						m_aEigVal[i]=dTmp;
						}
					}
									
				//-- copy Eigenvectors 
				Array2D< T > aEigVecTemp( iDim, iDim );
				oQEigVal.getV( aEigVecTemp );

				//-- sort/normalize  eigenvectors
				m_aEigVec = Array2D< T > ( iDim, iDim );
				for( j=0; j<iDim; ++j ) {
					dTmp=0.0;
					for( i=0; i<iDim; ++i ) 
						dTmp += sqrt( aEigVecTemp[i][aMap[j]] * aEigVecTemp[i][aMap[j]] );
					if( dTmp == 0.0 ) 
						dTmp = 1.0;
					for( i=0; i<iDim; ++i ) 
						m_aEigVec[i][j] = aEigVecTemp[i][aMap[j]] / dTmp ;
					}
				
				m_iStatus |= MDL_EIGENVAL;
				}
				
			//-----
			void CalcA( T dSampling ) { 
				int i, j, k, iDim = m_aQ.dim1();
			
				//-- aEVI = Inverse of aEigVec
				TNT::Array2D< T > aEVI = m_aEigVec.copy();
				MatrixInvert( aEVI );
				
				//----- A matrix  aa[k][i][j]=v[i][k]*vi[k][j]   and PEq
				//TNT::Array3D<double> AA( iDim, iDim, iDim );		// Spectral matrices ( if desired...) 
				m_aA = Array2D<T>( iDim, iDim, 0.0);
				double e_t;			// e ^ t 
				m_aPEq = Array1D< T > ( iDim );

				for( k=0; k<iDim; ++k ) {
					e_t = exp( -1.0 * m_aEigVal[k] * dSampling ) ;
					for( i=0; i<iDim; ++i ) 
						for( j=0; j<iDim; ++j ) 
							m_aA[i][j] += e_t * ( /* AA[k][i][j] = */ m_aEigVec[i][k] * aEVI[k][j] ) ;
					if( k==0 ) { 
						for( int iCl=0; iCl<iDim; ++iCl ) 
							m_aPEq[iCl]=m_aA[0][iCl];
						}
					}
				m_iStatus |= MDL_AMATRIX;
				}
				
		//----- Write class contents to debug window 
		#ifdef _DEBUG
		public : 
			void OutputDebug() { 
				OutputDebugArray( m_aQ, "Q   " );
				OutputDebugArray( m_aEigVec, "E Vec" );
				OutputDebugVector( m_aEigVal, "E Val" );
				OutputDebugArray( m_aA, "A   " );
				OutputDebugVector( m_aPEq, "PEq " );
				}
		#endif
			
		};
	
	}
#endif 
